##
# File: NmrPipePKParserListener.py
# Date: 03-Dec-2024
#
# Updates:
""" ParserLister class for NMRPIPE PK files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys
import re
import copy
import math

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.io.CifReader import CifReader
    from wwpdb.utils.nmr.pk.NmrPipePKParser import NmrPipePKParser
    from wwpdb.utils.nmr.pk.BasePKParserListener import BasePKParserListener
    from wwpdb.utils.nmr.mr.ParserListenerUtil import (ISOTOPE_NUMBERS_OF_NMR_OBS_NUCS,
                                                       REPRESENTATIVE_MODEL_ID,
                                                       REPRESENTATIVE_ALT_ID,
                                                       SPECTRAL_DIM_TEMPLATE,
                                                       getMaxEffDigits,
                                                       roundString)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
    from wwpdb.utils.nmr.AlignUtil import emptyValue
except ImportError:
    from nmr.io.CifReader import CifReader
    from nmr.pk.NmrPipePKParser import NmrPipePKParser
    from nmr.pk.BasePKParserListener import BasePKParserListener
    from nmr.mr.ParserListenerUtil import (ISOTOPE_NUMBERS_OF_NMR_OBS_NUCS,
                                           REPRESENTATIVE_MODEL_ID,
                                           REPRESENTATIVE_ALT_ID,
                                           SPECTRAL_DIM_TEMPLATE,
                                           getMaxEffDigits,
                                           roundString)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator
    from nmr.AlignUtil import emptyValue


# This class defines a complete listener for a parse tree produced by NmrPipePKParser.
class NmrPipePKParserListener(ParseTreeListener, BasePKParserListener):

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 representativeModelId: int = REPRESENTATIVE_MODEL_ID,
                 representativeAltId: str = REPRESENTATIVE_ALT_ID,
                 mrAtomNameMapping: Optional[List[dict]] = None,
                 cR: Optional[CifReader] = None, caC: Optional[dict] = None, ccU: Optional[ChemCompUtil] = None,
                 csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, representativeModelId, representativeAltId,
                         mrAtomNameMapping, cR, caC, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-pea-pip'
        self.software_name = 'NMRPIPE'

    # Enter a parse tree produced by NmrPipePKParser#nmrpipe_pk.
    def enterNmrpipe_pk(self, ctx: NmrPipePKParser.Nmrpipe_pkContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by NmrPipePKParser#dynamo_mr.
    def exitNmrpipe_pk(self, ctx: NmrPipePKParser.Nmrpipe_pkContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by NmrPipePKParser#data_label.
    def enterData_label(self, ctx: NmrPipePKParser.Data_labelContext):

        def set_spectral_dim(_dim_id):
            self.num_of_dim = max(self.num_of_dim, _dim_id)
            if _dim_id not in self.cur_spectral_dim:
                cur_spectral_dim = copy.copy(SPECTRAL_DIM_TEMPLATE)
            else:
                cur_spectral_dim = self.cur_spectral_dim[_dim_id]

            if ctx.Simple_name_DA():
                cur_spectral_dim['axis_code'] = axis_code = str(ctx.Simple_name_DA())
                if axis_code not in emptyValue:
                    digits = re.findall(r'\d+', axis_code)
                    for digit in digits:
                        num = int(digit)
                        nuc = next((k for k, v in ISOTOPE_NUMBERS_OF_NMR_OBS_NUCS.items() if num == v[0]), None)
                        if nuc is not None:
                            cur_spectral_dim['atom_type'] = nuc
                            cur_spectral_dim['atom_isotope_number'] = num
                            break
                    if cur_spectral_dim['atom_type'] is None:
                        for a in axis_code:
                            a = a.upper()
                            if a in ISOTOPE_NUMBERS_OF_NMR_OBS_NUCS:
                                cur_spectral_dim['atom_type'] = a
                                cur_spectral_dim['atom_isotope_number'] = ISOTOPE_NUMBERS_OF_NMR_OBS_NUCS[a][0]
                                break

            truncated = False
            first_point = last_point = None
            if ctx.Integer_DA(0) and ctx.Integer_DA(1):
                first_point = int(str(ctx.Integer_DA(0)))
                last_point = int(str(ctx.Integer_DA(1)))

                truncated |= first_point != 1
                log2_last_point = math.log2(float(last_point))
                round_log2_last_point = int(log2_last_point)
                truncated |= log2_last_point - round_log2_last_point > 0.001

            isotope_number = cur_spectral_dim['atom_isotope_number']

            cur_spectral_dim['acquisition'] = 'yes' if _dim_id == self.acq_dim_id\
                and (isotope_number == 1 or (isotope_number == 13 and self.exptlMethod == 'SOLID-STATE NMR')) else 'no'

            if _dim_id == 1 and cur_spectral_dim['acquisition'] == 'no':
                self.acq_dim_id = self.num_of_dim

            cur_spectral_dim['under_sampling_type'] = 'not observed' if cur_spectral_dim['acquisition'] == 'yes' else 'aliased'

            if ctx.Ppm_value_DA(0) and ctx.Ppm_value_DA(1):
                row = [str(ctx.Ppm_value_DA(0)).replace('ppm', ''),
                       str(ctx.Ppm_value_DA(1)).replace('ppm', '')]

                max_eff_digits = getMaxEffDigits(row)

                first_ppm = float(row[0])
                last_ppm = float(row[1])
                cur_spectral_dim['value_first_point'] = first_ppm
                cur_spectral_dim['center_frequency_offset'] = float(roundString(str((first_ppm + last_ppm) / 2.0),
                                                                                max_eff_digits))

                if truncated and isotope_number in (1, 13, 15):
                    _last_point = int(pow(2.0, round_log2_last_point + 1.0))
                    scale = float(_last_point) / float(last_point - first_point - 1)
                    center = first_ppm - (first_ppm - last_ppm) * scale / 2

                    if isotope_number == 1:
                        if not 4 < center < 6:
                            __last_point = _last_point + _last_point / 2
                            _scale = float(__last_point) / float(last_point - first_point - 1)
                            center = first_ppm - (first_ppm - last_ppm) * _scale / 2
                            if 4 < center < 6:
                                scale = _scale
                            else:
                                __last_point = _last_point + _last_point
                                _scale = float(__last_point) / float(last_point - first_point - 1)
                                center = first_ppm - (first_ppm - last_ppm) * _scale / 2
                                if 4 < center < 6:
                                    scale = _scale

                    cur_spectral_dim['sweep_width'] = float(roundString(str((first_ppm - last_ppm) * scale),
                                                                        max_eff_digits))
                else:
                    cur_spectral_dim['sweep_width'] = float(roundString(str(first_ppm - last_ppm),
                                                                        max_eff_digits))

                cur_spectral_dim['sweep_width_units'] = 'ppm'

            self.cur_spectral_dim[_dim_id] = cur_spectral_dim

        if ctx.X_axis_DA():
            set_spectral_dim(1)
        elif ctx.Y_axis_DA():
            set_spectral_dim(2)
        elif ctx.Z_axis_DA():
            set_spectral_dim(3)
        elif ctx.A_axis_DA():
            set_spectral_dim(4)

    # Exit a parse tree produced by NmrPipePKParser#data_label.
    def exitData_label(self, ctx: NmrPipePKParser.Data_labelContext):  # pylint: disable=unused-argument
        if self.num_of_dim > 2 and self.exptlMethod != 'SOLID-STATE NMR':
            for _dim_id in range(2, self.num_of_dim + 1):
                cur_spectral_dim = self.cur_spectral_dim[_dim_id]
                if cur_spectral_dim['atom_isotope_number'] != 13:
                    continue
                if cur_spectral_dim['center_frequency_offset'] > 100.0:
                    continue
                if cur_spectral_dim['sweep_width'] < 50.0:
                    cur_spectral_dim['under_sampling_type'] = 'folded'

    # Enter a parse tree produced by NmrPipePKParser#peak_list_2d.
    def enterPeak_list_2d(self, ctx: NmrPipePKParser.Peak_list_2dContext):
        self.initSpectralDim()

        self.null_value = str(ctx.Null_value()) if ctx.Null_value() else None
        self.null_string = str(ctx.Null_string()) if ctx.Null_string() else None
        self.software_name = 'NMRPIPE'

    # Exit a parse tree produced by NmrPipePKParser#peak_list_2d.
    def exitPeak_list_2d(self, ctx: NmrPipePKParser.Peak_list_2dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#peak_2d.
    def enterPeak_2d(self, ctx: NmrPipePKParser.Peak_2dContext):  # pylint: disable=unused-argument
        self.peaks2D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#peak_2d.
    def exitPeak_2d(self, ctx: NmrPipePKParser.Peak_2dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks2D -= 1
                return

            try:

                index = int(str(ctx.Integer(0)))
                # x_axis = self.numberSelection[0]
                # y_axis = self.numberSelection[1]
                # dx = self.numberSelection[2]
                # dy = self.numberSelection[3]
                x_ppm = self.numberSelection[4]
                y_ppm = self.numberSelection[5]
                x_hz = self.numberSelection[6]
                y_hz = self.numberSelection[7]
                # xw = self.numberSelection[8]
                # yw = self.numberSelection[9]
                xw_hz = self.numberSelection[10]
                yw_hz = self.numberSelection[11]
                # x1 = int(str(ctx.Integer(1)))
                # x3 = int(str(ctx.Integer(2)))
                # y1 = int(str(ctx.Integer(3)))
                # y3 = int(str(ctx.Integer(4)))
                height = self.originalNumberSelection[12]
                dheight = self.originalNumberSelection[13]
                vol = self.originalNumberSelection[14]
                # pchi2 = self.originalNumberSelection[15]
                # type = int(str(ctx.Integer(5)))

            except IndexError:
                self.peaks2D -= 1
                return

            if ctx.Any_name():
                ass = str(ctx.Any_name())
                if ass in emptyValue or (self.null_string is not None and ass == self.null_string):
                    ass = None
            else:
                ass = None
            # clustid = int(str(ctx.Integer(6)))
            # memcnt = int(str(ctx.Integer(7)))

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm):  # or type != 1:
                self.peaks2D -= 1
                return

            dstFunc = self.validatePeak2D(index, x_ppm, y_ppm, None, None, None, None,
                                          x_hz, y_hz, xw_hz, yw_hz, height, dheight, vol, None)

            if dstFunc is None:
                self.peaks2D -= 1
                return

            if self.num_of_dim != 2:
                self.num_of_dim = 2
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)

            if x_ppm is not None and x_hz is not None and cur_spectral_dim[1]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[1]['value_first_point'], float):
                cur_spectral_dim[1]['obs_freq_hint'].append(float(roundString(str(x_hz / (cur_spectral_dim[1]['value_first_point'] - x_ppm)),
                                                                              getMaxEffDigits([str(x_ppm), str(x_hz)]))))
            if y_ppm is not None and y_hz is not None and cur_spectral_dim[2]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[2]['value_first_point'], float):
                cur_spectral_dim[2]['obs_freq_hint'].append(float(roundString(str(y_hz / (cur_spectral_dim[2]['value_first_point'] - y_ppm)),
                                                                              getMaxEffDigits([str(y_ppm), str(y_hz)]))))

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = None

            if ass is not None:
                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2 =\
                    self.checkAssignments2D(index, assignments, dstFunc)

            self.addAssignedPkRow2D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    # Enter a parse tree produced by NmrPipePKParser#peak_list_3d.
    def enterPeak_list_3d(self, ctx: NmrPipePKParser.Peak_list_3dContext):
        self.initSpectralDim()

        self.null_value = str(ctx.Null_value()) if ctx.Null_value() else None
        self.null_string = str(ctx.Null_string()) if ctx.Null_string() else None
        self.software_name = 'NMRPIPE'

    # Exit a parse tree produced by NmrPipePKParser#peak_list_3d.
    def exitPeak_list_3d(self, ctx: NmrPipePKParser.Peak_list_3dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#peak_3d.
    def enterPeak_3d(self, ctx: NmrPipePKParser.Peak_3dContext):  # pylint: disable=unused-argument
        self.peaks3D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#peak_3d.
    def exitPeak_3d(self, ctx: NmrPipePKParser.Peak_3dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks3D -= 1
                return

            try:

                index = int(str(ctx.Integer(0)))
                # x_axis = self.numberSelection[0]
                # y_axis = self.numberSelection[1]
                # z_axis = self.numberSelection[2]
                # dx = self.numberSelection[3]
                # dy = self.numberSelection[4]
                # dz = self.numberSelection[5]
                x_ppm = self.numberSelection[6]
                y_ppm = self.numberSelection[7]
                z_ppm = self.numberSelection[8]
                x_hz = self.numberSelection[9]
                y_hz = self.numberSelection[10]
                z_hz = self.numberSelection[11]
                # xw = self.numberSelection[12]
                # yw = self.numberSelection[13]
                # zw = self.numberSelection[14]
                xw_hz = self.numberSelection[15]
                yw_hz = self.numberSelection[16]
                zw_hz = self.numberSelection[17]
                # x1 = int(str(ctx.Integer(1)))
                # x3 = int(str(ctx.Integer(2)))
                # y1 = int(str(ctx.Integer(3)))
                # y3 = int(str(ctx.Integer(4)))
                # z1 = int(str(ctx.Integer(5)))
                # z3 = int(str(ctx.Integer(6)))
                height = self.originalNumberSelection[18]
                dheight = self.originalNumberSelection[19]
                vol = self.originalNumberSelection[20]
                # pchi2 = self.originalNumberSelection[21]
                # type = int(str(ctx.Integer(7)))

            except IndexError:
                self.peaks3D -= 1
                return

            if ctx.Any_name():
                ass = str(ctx.Any_name())
                if ass in emptyValue or (self.null_string is not None and ass == self.null_string):
                    ass = None
            else:
                ass = None
            # clustid = int(str(ctx.Integer(8)))
            # memcnt = int(str(ctx.Integer(9)))

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm):  # or type != 1:
                self.peaks3D -= 1
                return

            dstFunc = self.validatePeak3D(index, x_ppm, y_ppm, z_ppm, None, None, None, None, None, None,
                                          x_hz, y_hz, z_hz, xw_hz, yw_hz, zw_hz, height, dheight, vol, None)

            if dstFunc is None:
                self.peaks3D -= 1
                return

            if self.num_of_dim != 3:
                self.num_of_dim = 3
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)

            if x_ppm is not None and x_hz is not None and cur_spectral_dim[1]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[1]['value_first_point'], float):
                cur_spectral_dim[1]['obs_freq_hint'].append(float(roundString(str(x_hz / (cur_spectral_dim[1]['value_first_point'] - x_ppm)),
                                                                              getMaxEffDigits([str(x_ppm), str(x_hz)]))))
            if y_ppm is not None and y_hz is not None and cur_spectral_dim[2]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[2]['value_first_point'], float):
                cur_spectral_dim[2]['obs_freq_hint'].append(float(roundString(str(y_hz / (cur_spectral_dim[2]['value_first_point'] - y_ppm)),
                                                                              getMaxEffDigits([str(y_ppm), str(y_hz)]))))
            if z_ppm is not None and z_hz is not None and cur_spectral_dim[3]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[3]['value_first_point'], float):
                cur_spectral_dim[3]['obs_freq_hint'].append(float(roundString(str(z_hz / (cur_spectral_dim[3]['value_first_point'] - z_ppm)),
                                                                              getMaxEffDigits([str(z_ppm), str(z_hz)]))))

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = None

            if ass is not None:
                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                    self.checkAssignments3D(index, assignments, dstFunc)

            self.addAssignedPkRow3D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    # Enter a parse tree produced by NmrPipePKParser#peak_list_4d.
    def enterPeak_list_4d(self, ctx: NmrPipePKParser.Peak_list_4dContext):
        self.initSpectralDim()

        self.null_value = str(ctx.Null_value()) if ctx.Null_value() else None
        self.null_string = str(ctx.Null_string()) if ctx.Null_string() else None
        self.software_name = 'NMRPIPE'

    # Exit a parse tree produced by NmrPipePKParser#peak_list_4d.
    def exitPeak_list_4d(self, ctx: NmrPipePKParser.Peak_list_4dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#peak_4d.
    def enterPeak_4d(self, ctx: NmrPipePKParser.Peak_4dContext):  # pylint: disable=unused-argument
        self.peaks4D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#peak_4d.
    def exitPeak_4d(self, ctx: NmrPipePKParser.Peak_4dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks4D -= 1
                return

            try:

                index = int(str(ctx.Integer(0)))
                # x_axis = self.numberSelection[0]
                # y_axis = self.numberSelection[1]
                # z_axis = self.numberSelection[2]
                # a_axis = self.numberSelection[3]
                # dx = self.numberSelection[4]
                # dy = self.numberSelection[5]
                # dz = self.numberSelection[6]
                # da = self.numberSelection[7]
                x_ppm = self.numberSelection[8]
                y_ppm = self.numberSelection[9]
                z_ppm = self.numberSelection[10]
                a_ppm = self.numberSelection[11]
                x_hz = self.numberSelection[12]
                y_hz = self.numberSelection[13]
                z_hz = self.numberSelection[14]
                a_hz = self.numberSelection[15]
                # xw = self.numberSelection[16]
                # yw = self.numberSelection[17]
                # zw = self.numberSelection[18]
                # aw = self.numberSelection[19]
                xw_hz = self.numberSelection[20]
                yw_hz = self.numberSelection[21]
                zw_hz = self.numberSelection[22]
                aw_hz = self.numberSelection[23]
                # x1 = int(str(ctx.Integer(1)))
                # x3 = int(str(ctx.Integer(2)))
                # y1 = int(str(ctx.Integer(3)))
                # y3 = int(str(ctx.Integer(4)))
                # z1 = int(str(ctx.Integer(5)))
                # z3 = int(str(ctx.Integer(6)))
                # a1 = int(str(ctx.Integer(7)))
                # a3 = int(str(ctx.Integer(8)))
                height = self.originalNumberSelection[24]
                dheight = self.originalNumberSelection[25]
                vol = self.originalNumberSelection[26]
                # pchi2 = self.originalNumberSelection[27]
                # type = int(str(ctx.Integer(9)))

            except IndexError:
                self.peaks4D -= 1
                return

            if ctx.Any_name():
                ass = str(ctx.Any_name())
                if ass in emptyValue or (self.null_string is not None and ass == self.null_string):
                    ass = None
            else:
                ass = None
            # clustid = int(str(ctx.Integer(10)))
            # memcnt = int(str(ctx.Integer(11)))

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm, a_ppm):  # or type != 1:
                self.peaks4D -= 1
                return

            dstFunc = self.validatePeak4D(index, x_ppm, y_ppm, z_ppm, a_ppm, None, None, None, None, None, None, None, None,
                                          x_hz, y_hz, z_hz, a_hz, xw_hz, yw_hz, zw_hz, aw_hz, height, dheight, vol, None)

            if dstFunc is None:
                self.peaks4D -= 1
                return

            if self.num_of_dim != 4:
                self.num_of_dim = 4
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)
            cur_spectral_dim[4]['freq_hint'].append(a_ppm)

            if x_ppm is not None and x_hz is not None and cur_spectral_dim[1]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[1]['value_first_point'], float):
                cur_spectral_dim[1]['obs_freq_hint'].append(float(roundString(str(x_hz / (cur_spectral_dim[1]['value_first_point'] - x_ppm)),
                                                                              getMaxEffDigits([str(x_ppm), str(x_hz)]))))
            if y_ppm is not None and y_hz is not None and cur_spectral_dim[2]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[2]['value_first_point'], float):
                cur_spectral_dim[2]['obs_freq_hint'].append(float(roundString(str(y_hz / (cur_spectral_dim[2]['value_first_point'] - y_ppm)),
                                                                              getMaxEffDigits([str(y_ppm), str(y_hz)]))))
            if z_ppm is not None and z_hz is not None and cur_spectral_dim[3]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[3]['value_first_point'], float):
                cur_spectral_dim[3]['obs_freq_hint'].append(float(roundString(str(z_hz / (cur_spectral_dim[3]['value_first_point'] - z_ppm)),
                                                                              getMaxEffDigits([str(z_ppm), str(z_hz)]))))
            if a_ppm is not None and a_hz is not None and cur_spectral_dim[4]['spectrometer_frequency'] is None\
               and isinstance(cur_spectral_dim[4]['value_first_point'], float):
                cur_spectral_dim[4]['obs_freq_hint'].append(float(roundString(str(a_hz / (cur_spectral_dim[4]['value_first_point'] - a_ppm)),
                                                                              getMaxEffDigits([str(a_ppm), str(a_hz)]))))

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = asis4 = None

            if ass is not None:
                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                    self.checkAssignments4D(index, assignments, dstFunc)

            self.addAssignedPkRow4D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3, asis4,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    # Enter a parse tree produced by NmrPipePKParser#pipp_label.
    def enterPipp_label(self, ctx: NmrPipePKParser.Pipp_labelContext):
        if ctx.Dim_count_DA():
            self.num_of_dim = int(str(ctx.Integer_DA()))

    # Exit a parse tree produced by NmrPipePKParser#pipp_label.
    def exitPipp_label(self, ctx: NmrPipePKParser.Pipp_labelContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#pipp_axis.
    def enterPipp_axis(self, ctx: NmrPipePKParser.Pipp_axisContext):

        def set_spectral_dim(_dim_id):
            if _dim_id not in self.cur_spectral_dim:
                cur_spectral_dim = copy.copy(SPECTRAL_DIM_TEMPLATE)
            else:
                cur_spectral_dim = self.cur_spectral_dim[_dim_id]

            sw = str(ctx.Float_DA(0))
            sf = str(ctx.Float_DA(1))
            first_value_point = str(ctx.Float_DA(2))

            max_eff_digits = getMaxEffDigits([sf, first_value_point])

            sw = float(sw)
            sf = float(sf)
            first_ppm = float(roundString(str(float(first_value_point) / sf), max_eff_digits))
            last_ppm = float(roundString(str(first_ppm - sw / sf), max_eff_digits))
            center_ppm = (first_ppm + last_ppm) / 2.0

            cur_spectral_dim['spectrometer_frequency'] = sf
            cur_spectral_dim['sweep_width'] = sw
            cur_spectral_dim['sweep_width_units'] = 'Hz'
            cur_spectral_dim['value_first_point'] = first_ppm
            cur_spectral_dim['center_frequency_offset'] = float(roundString(str(center_ppm), max_eff_digits))

            self.cur_spectral_dim[_dim_id] = cur_spectral_dim

        if ctx.X_axis_DA():
            set_spectral_dim(1)
        elif ctx.Y_axis_DA():
            set_spectral_dim(2)
        elif ctx.Z_axis_DA():
            set_spectral_dim(3)
        elif ctx.A_axis_DA():
            set_spectral_dim(4)

    # Exit a parse tree produced by NmrPipePKParser#pipp_axis.
    def exitPipp_axis(self, ctx: NmrPipePKParser.Pipp_axisContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_list_2d.
    def enterPipp_peak_list_2d(self, ctx: NmrPipePKParser.Pipp_peak_list_2dContext):  # pylint: disable=unused-argument
        if self.num_of_dim != 2:
            self.num_of_dim = 2
        self.initSpectralDim()
        self.software_name = 'PIPP'

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_list_2d.
    def exitPipp_peak_list_2d(self, ctx: NmrPipePKParser.Pipp_peak_list_2dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_2d.
    def enterPipp_peak_2d(self, ctx: NmrPipePKParser.Pipp_peak_2dContext):  # pylint: disable=unused-argument
        self.peaks2D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_2d.
    def exitPipp_peak_2d(self, ctx: NmrPipePKParser.Pipp_peak_2dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks2D -= 1
                return

            try:

                index = int(str(ctx.Integer()))
                x_ppm = self.numberSelection[0]
                y_ppm = self.numberSelection[1]
                height = self.originalNumberSelection[2]

            except IndexError:
                self.peaks2D -= 1
                return

            if len(self.originalNumberSelection) > 3:
                ass1 = self.originalNumberSelection[3]
                if ass1 in emptyValue:
                    ass1 = None
            else:
                ass1 = None

            if len(self.originalNumberSelection) > 4:
                ass2 = self.originalNumberSelection[4]
                if ass2 in emptyValue:
                    ass2 = None
            else:
                ass2 = None

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm):
                self.peaks2D -= 1
                return

            dstFunc = self.validatePeak2D(index, x_ppm, y_ppm, None, None, None, None,
                                          None, None, None, None, height, None, None, None)

            if dstFunc is None:
                self.peaks2D -= 1
                return

            if self.num_of_dim != 2:
                self.num_of_dim = 2
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = None

            ass = None
            if ass1 is not None and ass2 is not None:
                ass = f'{ass1}, {ass2}'
            elif ass1 is not None:
                ass = ass1
            elif ass2 is not None:
                ass = ass2

            if ass is not None:
                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2 =\
                    self.checkAssignments2D(index, assignments, dstFunc)

            self.addAssignedPkRow2D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_list_3d.
    def enterPipp_peak_list_3d(self, ctx: NmrPipePKParser.Pipp_peak_list_3dContext):  # pylint: disable=unused-argument
        if self.num_of_dim != 3:
            self.num_of_dim = 3
        self.initSpectralDim()
        self.software_name = 'PIPP'

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_list_3d.
    def exitPipp_peak_list_3d(self, ctx: NmrPipePKParser.Pipp_peak_list_3dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_3d.
    def enterPipp_peak_3d(self, ctx: NmrPipePKParser.Pipp_peak_3dContext):  # pylint: disable=unused-argument
        self.peaks3D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_3d.
    def exitPipp_peak_3d(self, ctx: NmrPipePKParser.Pipp_peak_3dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks3D -= 1
                return

            try:

                index = int(str(ctx.Integer(0)))
                x_ppm = self.numberSelection[0]
                y_ppm = self.numberSelection[1]
                z_ppm = self.numberSelection[2]
                height = self.originalNumberSelection[3]

            except IndexError:
                self.peaks3D -= 1
                return

            if len(self.originalNumberSelection) > 4:
                ass1 = self.originalNumberSelection[4]
                if ass1 in emptyValue:
                    ass1 = None
            else:
                ass1 = None

            if len(self.originalNumberSelection) > 5:
                ass2 = self.originalNumberSelection[5]
                if ass2 in emptyValue:
                    ass2 = None
            else:
                ass2 = None

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm):
                self.peaks3D -= 1
                return

            dstFunc = self.validatePeak3D(index, x_ppm, y_ppm, z_ppm, None, None, None, None, None, None,
                                          None, None, None, None, None, None, height, None, None, None)

            if dstFunc is None:
                self.peaks3D -= 1
                return

            if self.num_of_dim != 3:
                self.num_of_dim = 3
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = None

            ass = None
            if ass1 is not None and ass2 is not None:
                ass = f'{ass1}, {ass2}'
            elif ass1 is not None:
                ass = ass1
            elif ass2 is not None:
                ass = ass2

            if ass is not None and self.reasons is not None and 'onebond_resolved' in self.reasons:

                if self.extractPeakAssignment(2, ass1, index) is not None:
                    onebondOrder = 0
                else:
                    onebondOrder = 1

                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                    self.checkAssignments3D(index, assignments, dstFunc, onebondOrder)

            self.addAssignedPkRow3D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_list_4d.
    def enterPipp_peak_list_4d(self, ctx: NmrPipePKParser.Pipp_peak_list_4dContext):  # pylint: disable=unused-argument
        if self.num_of_dim != 4:
            self.num_of_dim = 4
        self.initSpectralDim()
        self.software_name = 'PIPP'

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_list_4d.
    def exitPipp_peak_list_4d(self, ctx: NmrPipePKParser.Pipp_peak_list_4dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipePKParser#pipp_peak_4d.
    def enterPipp_peak_4d(self, ctx: NmrPipePKParser.Pipp_peak_4dContext):  # pylint: disable=unused-argument
        self.peaks4D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by NmrPipePKParser#pipp_peak_4d.
    def exitPipp_peak_4d(self, ctx: NmrPipePKParser.Pipp_peak_4dContext):

        try:

            if len(self.numberSelection) == 0:
                self.peaks4D -= 1
                return

            try:

                index = int(str(ctx.Integer(0)))
                x_ppm = self.numberSelection[0]
                y_ppm = self.numberSelection[1]
                z_ppm = self.numberSelection[2]
                a_ppm = self.numberSelection[3]
                height = self.originalNumberSelection[4]

            except IndexError:
                self.peaks4D -= 1
                return

            if len(self.originalNumberSelection) > 5:
                ass1 = self.originalNumberSelection[5]
                if ass1 in emptyValue:
                    ass1 = None
            else:
                ass1 = None

            if len(self.originalNumberSelection) > 6:
                ass2 = self.originalNumberSelection[6]
                if ass2 in emptyValue:
                    ass2 = None
            else:
                ass2 = None

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm, a_ppm):
                self.peaks4D -= 1
                return

            dstFunc = self.validatePeak4D(index, x_ppm, y_ppm, z_ppm, a_ppm, None, None, None, None, None, None, None, None,
                                          None, None, None, None, None, None, None, None, height, None, None, None)

            if dstFunc is None:
                self.peaks4D -= 1
                return

            if self.num_of_dim != 4:
                self.num_of_dim = 4
                self.initSpectralDim()

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)
            cur_spectral_dim[4]['freq_hint'].append(a_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = asis4 = None

            ass = None
            if ass1 is not None and ass2 is not None:
                ass = f'{ass1}, {ass2}'
            elif ass1 is not None:
                ass = ass1
            elif ass2 is not None:
                ass = ass2

            if ass is not None and self.reasons is not None and 'onebond_resolved' in self.reasons:
                assignments = []
                if len(ass.split('-')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(':')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(':'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(';')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(';'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                elif len(ass.split(',')) == self.num_of_dim:
                    hint = None
                    for _ass in ass.split(','):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None
                else:
                    assignments = [None] * self.num_of_dim
                    _assignments = self.extractPeakAssignment(self.num_of_dim, ass, index)
                    if _assignments is not None and len(_assignments) == self.num_of_dim:
                        for idx in range(self.num_of_dim):
                            assignments[idx] = [_assignments[idx]]  # pylint: disable=unsubscriptable-object

                has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                    self.checkAssignments4D(index, assignments, dstFunc)

            self.addAssignedPkRow4D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3, asis4,
                                    f'{ass} -> ',
                                    None if has_assignments and not has_multiple_assignments else ass)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()

    def enterNumber(self, ctx: NmrPipePKParser.NumberContext):  # pylint: disable=unused-argument

        try:

            if ctx.Integer():
                value = str(ctx.Integer())
                if self.null_value is not None and value == self.null_value:
                    self.numberSelection.append(None)
                    self.originalNumberSelection.append(None)
                else:
                    self.numberSelection.append(int(value))
                    self.originalNumberSelection.append(value)

            elif ctx.Float():
                value = str(ctx.Float())
                if self.null_value is not None and value == self.null_value:
                    self.numberSelection.append(None)
                    self.originalNumberSelection.append(None)
                else:
                    self.numberSelection.append(float(value))
                    self.originalNumberSelection.append(value)

            elif ctx.Real():
                value = str(ctx.Real())
                if self.null_value is not None and value == self.null_value:
                    self.numberSelection.append(None)
                    self.originalNumberSelection.append(None)
                else:
                    self.numberSelection.append(float(value))
                    self.originalNumberSelection.append(value)

            else:
                self.numberSelection.append(None)
                self.originalNumberSelection.append(str(ctx.Any_name()))

        except ValueError:
            self.numberSelection.append(None)
            self.originalNumberSelection.append(None)

    # Exit a parse tree produced by NmrPipePKParser#number.
    def exitNumber(self, ctx: NmrPipePKParser.NumberContext):  # pylint: disable=unused-argument
        pass


# del NmrPipePKParser



##
# File: CcpnPKParserListener.py
# Date: 26-Feb-2025
#
# Updates:
""" ParserLister class for CCPN PK files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.io.CifReader import CifReader
    from wwpdb.utils.nmr.pk.CcpnPKParser import CcpnPKParser
    from wwpdb.utils.nmr.pk.BasePKParserListener import BasePKParserListener
    from wwpdb.utils.nmr.mr.ParserListenerUtil import (REPRESENTATIVE_MODEL_ID,
                                                       REPRESENTATIVE_ALT_ID)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
    from wwpdb.utils.nmr.AlignUtil import emptyValue
except ImportError:
    from nmr.io.CifReader import CifReader
    from nmr.pk.CcpnPKParser import CcpnPKParser
    from nmr.pk.BasePKParserListener import BasePKParserListener
    from nmr.mr.ParserListenerUtil import (REPRESENTATIVE_MODEL_ID,
                                           REPRESENTATIVE_ALT_ID)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator
    from nmr.AlignUtil import emptyValue


# This class defines a complete listener for a parse tree produced by CcpnPKParser.
class CcpnPKParserListener(ParseTreeListener, BasePKParserListener):

    __has_number = False
    __has_id = False
    __has_height = False
    __has_volume = False
    __has_lw_hz = False
    __has_merit = False
    __has_details = False

    __noteSelection = []

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 representativeModelId: int = REPRESENTATIVE_MODEL_ID,
                 representativeAltId: str = REPRESENTATIVE_ALT_ID,
                 mrAtomNameMapping: Optional[List[dict]] = None,
                 cR: Optional[CifReader] = None, caC: Optional[dict] = None, ccU: Optional[ChemCompUtil] = None,
                 csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, representativeModelId, representativeAltId,
                         mrAtomNameMapping, cR, caC, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-pea-ccp'
        self.software_name = 'CCPN'

    # Enter a parse tree produced by CcpnPKParser#ccpn_pk.
    def enterCcpn_pk(self, ctx: CcpnPKParser.Ccpn_pkContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by CcpnPKParser#ccpn_pk.
    def exitCcpn_pk(self, ctx: CcpnPKParser.Ccpn_pkContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by CcpnPKParser#peak_list_2d.
    def enterPeak_list_2d(self, ctx: CcpnPKParser.Peak_list_2dContext):
        self.num_of_dim = 2
        self.initSpectralDim()

        self.__has_number = False
        if ctx.Num():
            self.__has_number = True

        self.__has_id = False
        if ctx.Id() or ctx.Id_():
            self.__has_id = True

        self.__has_height = False
        if ctx.Height():
            self.__has_height = True

        self.__has_volume = False
        if ctx.Volume():
            self.__has_volume = True

        self.__has_lw_hz = False
        if ctx.Line_width_F1():
            self.__has_lw_hz = True

        self.__has_merit = False
        if ctx.Merit():
            self.__has_merit = True

        self.__has_details = False
        if ctx.Details():
            self.__has_details = True

    # Exit a parse tree produced by CcpnPKParser#peak_list_2d.
    def exitPeak_list_2d(self, ctx: CcpnPKParser.Peak_list_2dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by CcpnPKParser#peak_2d.
    def enterPeak_2d(self, ctx: CcpnPKParser.Peak_2dContext):  # pylint: disable=unused-argument
        self.peaks2D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by CcpnPKParser#peak_2d.
    def exitPeak_2d(self, ctx: CcpnPKParser.Peak_2dContext):

        try:

            if 0 in (len(self.positionSelection), len(self.numberSelection)):
                self.peaks2D -= 1
                return

            if self.__has_id:
                index = int(str(ctx.Integer(1 if self.__has_number else 0)))
            elif self.__has_number:
                index = int(str(ctx.Integer(0)))
            else:
                index = self.peaks2D

            float_offset = 0

            try:

                P1 = self.positionSelection[float_offset]
                P2 = self.positionSelection[float_offset + 1]
                float_offset += 2

                W1 = W2 = merit = None

                if self.__has_lw_hz:
                    W1 = self.positionSelection[float_offset]
                    W2 = self.positionSelection[float_offset + 1]
                    float_offset += 2

                if self.__has_merit:
                    merit = self.positionSelection[float_offset]

            except IndexError:
                self.peaks2D -= 1
                return

            L1 = str(ctx.Simple_name(0))
            L2 = str(ctx.Simple_name(1))

            if L1 in emptyValue:
                L1 = None
            if L2 in emptyValue:
                L2 = None

            height = self.originalNumberSelection[0] if self.__has_height else None
            volume = self.originalNumberSelection[1 if self.__has_height else 0] if self.__has_volume else None

            details = None
            if self.__has_details:
                details = self.__noteSelection[0]
                if details in emptyValue:
                    details = None

            # fit_method = self.__noteSelection[1]
            # vol_method = self.__noteSelection[2]
            # if len(self.__noteSelection) > 3:
            #     vol_method += ' ' + self.__noteSelection[3]

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (P1, P2):
                self.peaks2D -= 1
                return

            dstFunc = self.validatePeak2D(index, P1, P2, None, None, None, None,
                                          None, None, W1, W2, height, None, volume, None, merit)

            if dstFunc is None:
                self.peaks2D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(P1)
            cur_spectral_dim[2]['freq_hint'].append(P2)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = None

            if None not in (L1, L2):
                assignments = [None] * self.num_of_dim
                if L1 is not None:
                    ext = self.extractPeakAssignment(1, L1, index)
                    if ext is not None:
                        status, _L1 = self.testAssignment(1, ext, L1)
                        if status:
                            assignments[0] = ext
                        elif _L1 is not None:
                            if details is None:
                                details = f'Assign 1:{L1} -> {_L1}'
                            ext = self.extractPeakAssignment(1, _L1, index)
                            if ext is not None:
                                assignments[0] = ext
                if L2 is not None:
                    ext = self.extractPeakAssignment(1, L2, index)
                    if ext is not None:
                        status, _L2 = self.testAssignment(2, ext, L2)
                        if status:
                            assignments[1] = ext
                        elif _L2 is not None:
                            if details is None:
                                details = f'Assign 2:{L2} -> {_L2}'
                            ext = self.extractPeakAssignment(1, _L2, index)
                            if ext is not None:
                                assignments[1] = ext

                has_assignments, has_multiple_assignments, asis1, asis2 =\
                    self.checkAssignments2D(index, assignments, dstFunc)

            self.addAssignedPkRow2D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2,
                                    f'{L1} {L2} -> ',
                                    details if details is not None or None in (L1, L2) or (has_assignments and not has_multiple_assignments)
                                    else f'{L1} {L2}')

        finally:
            self.positionSelection.clear()
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__noteSelection.clear()

    # Enter a parse tree produced by CcpnPKParser#peak_list_3d.
    def enterPeak_list_3d(self, ctx: CcpnPKParser.Peak_list_3dContext):
        self.num_of_dim = 3
        self.initSpectralDim()

        self.__has_number = False
        if ctx.Num():
            self.__has_number = True

        self.__has_id = False
        if ctx.Id() or ctx.Id_():
            self.__has_id = True

        self.__has_height = False
        if ctx.Height():
            self.__has_height = True

        self.__has_volume = False
        if ctx.Volume():
            self.__has_volume = True

        self.__has_lw_hz = False
        if ctx.Line_width_F1():
            self.__has_lw_hz = True

        self.__has_merit = False
        if ctx.Merit():
            self.__has_merit = True

        self.__has_details = False
        if ctx.Details():
            self.__has_details = True

    # Exit a parse tree produced by CcpnPKParser#peak_list_3d.
    def exitPeak_list_3d(self, ctx: CcpnPKParser.Peak_list_3dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by CcpnPKParser#peak_3d.
    def enterPeak_3d(self, ctx: CcpnPKParser.Peak_3dContext):  # pylint: disable=unused-argument
        self.peaks3D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by CcpnPKParser#peak_3d.
    def exitPeak_3d(self, ctx: CcpnPKParser.Peak_3dContext):

        try:

            if 0 in (len(self.positionSelection), len(self.numberSelection)):
                self.peaks3D -= 1
                return

            if self.__has_id:
                index = int(str(ctx.Integer(1 if self.__has_number else 0)))
            elif self.__has_number:
                index = int(str(ctx.Integer(0)))
            else:
                index = self.peaks3D

            float_offset = 0

            try:

                P1 = self.positionSelection[float_offset]
                P2 = self.positionSelection[float_offset + 1]
                P3 = self.positionSelection[float_offset + 2]
                float_offset += 3

                W1 = W2 = W3 = merit = None

                if self.__has_lw_hz:
                    W1 = self.positionSelection[float_offset]
                    W2 = self.positionSelection[float_offset + 1]
                    W3 = self.positionSelection[float_offset + 2]
                    float_offset += 3

                if self.__has_merit:
                    merit = self.positionSelection[float_offset]

            except IndexError:
                self.peaks3D -= 1
                return

            L1 = str(ctx.Simple_name(0))
            L2 = str(ctx.Simple_name(1))
            L3 = str(ctx.Simple_name(2))

            if L1 in emptyValue:
                L1 = None
            if L2 in emptyValue:
                L2 = None
            if L3 in emptyValue:
                L3 = None

            height = self.originalNumberSelection[0] if self.__has_height else None
            volume = self.originalNumberSelection[1 if self.__has_height else 0] if self.__has_volume else None

            details = None
            if self.__has_details:
                details = self.__noteSelection[0]
                if details in emptyValue:
                    details = None

            # fit_method = self.__noteSelection[1]
            # vol_method = self.__noteSelection[2]
            # if len(self.__noteSelection) > 3:
            #     vol_method += ' ' + self.__noteSelection[3]

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (P1, P2, P3):
                self.peaks3D -= 1
                return

            dstFunc = self.validatePeak3D(index, P1, P2, P3, None, None, None, None, None, None,
                                          None, None, None, W1, W2, W3, height, None, volume, None, merit)

            if dstFunc is None:
                self.peaks3D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(P1)
            cur_spectral_dim[2]['freq_hint'].append(P2)
            cur_spectral_dim[3]['freq_hint'].append(P3)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = None

            if None not in (L1, L2, L3):
                assignments = [None] * self.num_of_dim
                if L1 is not None:
                    ext = self.extractPeakAssignment(1, L1, index)
                    if ext is not None:
                        status, _L1 = self.testAssignment(1, ext, L1)
                        if status:
                            assignments[0] = ext
                        elif _L1 is not None:
                            if details is None:
                                details = f'Assign 1:{L1} -> {_L1}'
                            ext = self.extractPeakAssignment(1, _L1, index)
                            if ext is not None:
                                assignments[0] = ext
                if L2 is not None:
                    ext = self.extractPeakAssignment(1, L2, index)
                    if ext is not None:
                        status, _L2 = self.testAssignment(2, ext, L2)
                        if status:
                            assignments[1] = ext
                        elif _L2 is not None:
                            if details is None:
                                details = f'Assign 2:{L2} -> {_L2}'
                            ext = self.extractPeakAssignment(1, _L2, index)
                            if ext is not None:
                                assignments[1] = ext
                if L3 is not None:
                    ext = self.extractPeakAssignment(1, L3, index)
                    if ext is not None:
                        status, _L3 = self.testAssignment(3, ext, L3)
                        if status:
                            assignments[2] = ext
                        elif _L3 is not None:
                            if details is None:
                                details = f'Assign 3:{L3} -> {_L3}'
                            ext = self.extractPeakAssignment(1, _L3, index)
                            if ext is not None:
                                assignments[2] = ext

                has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                    self.checkAssignments3D(index, assignments, dstFunc)

            self.addAssignedPkRow3D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3,
                                    f'{L1} {L2} {L3} -> ',
                                    details if details is not None or None in (L1, L2, L3) or (has_assignments and not has_multiple_assignments)
                                    else f'{L1} {L2} {L3}')

        finally:
            self.positionSelection.clear()
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__noteSelection.clear()

    # Enter a parse tree produced by CcpnPKParser#peak_list_4d.
    def enterPeak_list_4d(self, ctx: CcpnPKParser.Peak_list_4dContext):
        self.num_of_dim = 4
        self.initSpectralDim()

        self.__has_number = False
        if ctx.Num():
            self.__has_number = True

        self.__has_id = False
        if ctx.Id() or ctx.Id_():
            self.__has_id = True

        self.__has_height = False
        if ctx.Height():
            self.__has_height = True

        self.__has_volume = False
        if ctx.Volume():
            self.__has_volume = True

        self.__has_lw_hz = False
        if ctx.Line_width_F1():
            self.__has_lw_hz = True

        self.__has_merit = False
        if ctx.Merit():
            self.__has_merit = True

        self.__has_details = False
        if ctx.Details():
            self.__has_details = True

    # Exit a parse tree produced by CcpnPKParser#peak_list_4d.
    def exitPeak_list_4d(self, ctx: CcpnPKParser.Peak_list_4dContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by CcpnPKParser#peak_4d.
    def enterPeak_4d(self, ctx: CcpnPKParser.Peak_4dContext):  # pylint: disable=unused-argument
        self.peaks4D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by CcpnPKParser#peak_4d.
    def exitPeak_4d(self, ctx: CcpnPKParser.Peak_4dContext):

        try:

            if 0 in (len(self.positionSelection), len(self.numberSelection)):
                self.peaks4D -= 1
                return

            if self.__has_id:
                index = int(str(ctx.Integer(1 if self.__has_number else 0)))
            elif self.__has_number:
                index = int(str(ctx.Integer(0)))
            else:
                index = self.peaks4D

            float_offset = 0

            try:

                P1 = self.positionSelection[float_offset]
                P2 = self.positionSelection[float_offset + 1]
                P3 = self.positionSelection[float_offset + 2]
                P4 = self.positionSelection[float_offset + 3]
                float_offset += 4

                W1 = W2 = W3 = W4 = merit = None

                if self.__has_lw_hz:
                    W1 = self.positionSelection[float_offset]
                    W2 = self.positionSelection[float_offset + 1]
                    W3 = self.positionSelection[float_offset + 2]
                    W4 = self.positionSelection[float_offset + 3]
                    float_offset += 4

                if self.__has_merit:
                    merit = self.positionSelection[float_offset]

            except IndexError:
                self.peaks4D -= 1
                return

            L1 = str(ctx.Simple_name(0))
            L2 = str(ctx.Simple_name(1))
            L3 = str(ctx.Simple_name(2))
            L4 = str(ctx.Simple_name(3))

            if L1 in emptyValue:
                L1 = None
            if L2 in emptyValue:
                L2 = None
            if L3 in emptyValue:
                L3 = None
            if L4 in emptyValue:
                L4 = None

            height = self.originalNumberSelection[0] if self.__has_height else None
            volume = self.originalNumberSelection[1 if self.__has_height else 0] if self.__has_volume else None

            details = None
            if self.__has_details:
                details = self.__noteSelection[0]
                if details in emptyValue:
                    details = None

            # fit_method = self.__noteSelection[1]
            # vol_method = self.__noteSelection[2]
            # if len(self.__noteSelection) > 3:
            #     vol_method += ' ' + self.__noteSelection[3]

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (P1, P2, P3, P4):
                self.peaks4D -= 1
                return

            dstFunc = self.validatePeak4D(index, P1, P2, P3, P4, None, None, None, None, None, None, None, None,
                                          None, None, None, None, W1, W2, W3, W4, height, None, volume, None, merit)

            if dstFunc is None:
                self.peaks4D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(P1)
            cur_spectral_dim[2]['freq_hint'].append(P2)
            cur_spectral_dim[3]['freq_hint'].append(P3)
            cur_spectral_dim[4]['freq_hint'].append(P4)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = asis4 = None

            if None not in (L1, L2, L3, L4):
                assignments = [None] * self.num_of_dim
                if L1 is not None:
                    ext = self.extractPeakAssignment(1, L1, index)
                    if ext is not None:
                        status, _L1 = self.testAssignment(1, ext, L1)
                        if status:
                            assignments[0] = ext
                        elif _L1 is not None:
                            if details is None:
                                details = f'Assign 1:{L1} -> {_L1}'
                            ext = self.extractPeakAssignment(1, _L1, index)
                            if ext is not None:
                                assignments[0] = ext
                if L2 is not None:
                    ext = self.extractPeakAssignment(1, L2, index)
                    if ext is not None:
                        status, _L2 = self.testAssignment(2, ext, L2)
                        if status:
                            assignments[1] = ext
                        elif _L2 is not None:
                            if details is None:
                                details = f'Assign 2:{L2} -> {_L2}'
                            ext = self.extractPeakAssignment(1, _L2, index)
                            if ext is not None:
                                assignments[1] = ext
                if L3 is not None:
                    ext = self.extractPeakAssignment(1, L3, index)
                    if ext is not None:
                        status, _L3 = self.testAssignment(3, ext, L3)
                        if status:
                            assignments[2] = ext
                        elif _L3 is not None:
                            if details is None:
                                details = f'Assign 3:{L3} -> {_L3}'
                            ext = self.extractPeakAssignment(1, _L3, index)
                            if ext is not None:
                                assignments[2] = ext
                if L4 is not None:
                    ext = self.extractPeakAssignment(1, L4, index)
                    if ext is not None:
                        status, _L4 = self.testAssignment(4, ext, L4)
                        if status:
                            assignments[3] = ext
                        elif _L4 is not None:
                            if details is None:
                                details = f'Assign 4:{L4} -> {_L4}'
                            ext = self.extractPeakAssignment(1, _L4, index)
                            if ext is not None:
                                assignments[3] = ext

                has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                    self.checkAssignments4D(index, assignments, dstFunc)

            self.addAssignedPkRow4D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3, asis4,
                                    f'{L1} {L2} {L3} {L4} -> ',
                                    details if details is not None or None in (L1, L2, L3, L4) or (has_assignments and not has_multiple_assignments)
                                    else f'{L1} {L2} {L3} {L4}')

        finally:
            self.positionSelection.clear()
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__noteSelection.clear()

    # Enter a parse tree produced by CcpnPKParser#position.
    def enterPosition(self, ctx: CcpnPKParser.PositionContext):

        try:

            if ctx.Float():
                value = str(ctx.Float())
                self.positionSelection.append(float(value))

            elif ctx.Real():
                value = str(ctx.Real())
                self.positionSelection.append(float(value))

            elif ctx.Integer():
                value = str(ctx.Integer())
                self.positionSelection.append(float(value))

            else:
                self.positionSelection.append(None)

        except ValueError:
            self.positionSelection.append(None)

    # Exit a parse tree produced by CcpnPKParser#position.
    def exitPosition(self, ctx: CcpnPKParser.PositionContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by CcpnPKParser#number.
    def enterNumber(self, ctx: CcpnPKParser.NumberContext):

        try:

            if ctx.Float():
                value = str(ctx.Float())
                self.numberSelection.append(float(value))
                self.originalNumberSelection.append(value)

            elif ctx.Real():
                value = str(ctx.Real())
                self.numberSelection.append(float(value))
                self.originalNumberSelection.append(value)

            elif ctx.Integer():
                value = str(ctx.Integer())
                self.numberSelection.append(int(value))
                self.originalNumberSelection.append(value)

            else:
                self.numberSelection.append(None)
                self.originalNumberSelection.append(None)

        except ValueError:
            self.numberSelection.append(None)
            self.originalNumberSelection.append(None)

    # Exit a parse tree produced by CcpnPKParser#number.
    def exitNumber(self, ctx: CcpnPKParser.NumberContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by CcpnPKParser#note.
    def enterNote(self, ctx: CcpnPKParser.NoteContext):
        if ctx.Float():
            self.__noteSelection.append(str(ctx.Float()))

        elif ctx.Real():
            self.__noteSelection.append(str(ctx.Real()))

        elif ctx.Integer():
            self.__noteSelection.append(str(ctx.Integer()))

        elif ctx.Simple_name():
            self.__noteSelection.append(str(ctx.Simple_name()))

        else:
            self.__noteSelection.append(str(ctx.Any_name()))

    # Exit a parse tree produced by CcpnPKParser#note.
    def exitNote(self, ctx: CcpnPKParser.NoteContext):  # pylint: disable=unused-argument
        pass


# del CcpnPKParser



##
# File: OliviaCSParserListener.py
# Date: 22-Aug-2025
#
# Updates:
""" ParserLister class for OLIVIA CS files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.cs.OliviaCSParser import OliviaCSParser
    from wwpdb.utils.nmr.cs.BaseCSParserListener import BaseCSParserListener
    from wwpdb.utils.nmr.AlignUtil import (emptyValue, monDict3)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.cs.OliviaCSParser import OliviaCSParser
    from nmr.cs.BaseCSParserListener import BaseCSParserListener
    from nmr.AlignUtil import (emptyValue, monDict3)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator


# This class defines a complete listener for a parse tree produced by OliviaCSParser.
class OliviaCSParserListener(ParseTreeListener, BaseCSParserListener):

    __polySeq = None
    __entityAssembly = None

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 polySeq: List[dict] = None, entityAssembly: Optional[dict] = None,
                 ccU: Optional[ChemCompUtil] = None, csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, polySeq, entityAssembly, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-shi-oli'
        self.software_name = 'Olivia'

    # Enter a parse tree produced by OliviaCSParser#olivia_cs.
    def enterOlivia_cs(self, ctx: OliviaCSParser.Olivia_csContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by OliviaCSParser#olivia_cs.
    def exitOlivia_cs(self, ctx: OliviaCSParser.Olivia_csContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by OliviaCSParser#sequence.
    def enterSequence(self, ctx: OliviaCSParser.SequenceContext):  # pylint: disable=unused-argument
        self.__polySeq = []
        self.__entityAssembly = {}
        self.offset = {}

    # Exit a parse tree produced by OliviaCSParser#sequence.
    def exitSequence(self, ctx: OliviaCSParser.SequenceContext):  # pylint: disable=unused-argument

        if not self.hasPolySeq:

            self.polySeq = self.__polySeq
            self.entityAssembly = self.__entityAssembly
            self.hasPolySeq = True

            self.labelToAuthSeq = {}
            for ps in self.polySeq:
                chainId = ps['chain_id']
                for seqId in ps['seq_id']:
                    self.labelToAuthSeq[(chainId, seqId)] = (chainId, seqId)
            self.authToLabelSeq = {v: k for k, v in self.labelToAuthSeq.items()}

            self.chainIdSet = set(ps['chain_id'] for ps in self.polySeq)
            self.compIdSet = set()

            def is_data(array: list) -> bool:
                return not any(d in emptyValue for d in array)

            for ps in self.polySeq:
                self.compIdSet.update(set(filter(is_data, ps['comp_id'])))

            for compId in self.compIdSet:
                if compId in monDict3:
                    if len(compId) == 3:
                        self.polyPeptide = True
                    elif len(compId) == 2 and compId.startswith('D'):
                        self.polyDeoxyribonucleotide = True
                    elif len(compId) == 1:
                        self.polyRibonucleotide = True

    # Enter a parse tree produced by OliviaCSParser#residue.
    def enterResidue(self, ctx: OliviaCSParser.ResidueContext):

        chain_id = str(ctx.Simple_name(0))
        comp_id = str(ctx.Simple_name(1))
        seq_id = int(str(ctx.Integer()))

        ps = next((ps for ps in self.__polySeq if ps['chain_id'] == chain_id), None)

        if ps is None:
            self.__polySeq.append({'chain_id': chain_id, 'seq_id': [], 'comp_id': []})
            ps = self.__polySeq[-1]
            entity_assembly_id = len(self.__polySeq)
            self.__entityAssembly[str(entity_assembly_id)] = {'entity_id': entity_assembly_id, 'auth_asym_id': chain_id}
            self.predictSequenceNumberOffsetByFirstResidue(chain_id, seq_id, comp_id)

        ps['seq_id'].append(seq_id)
        ps['comp_id'].append(comp_id)

    # Exit a parse tree produced by OliviaCSParser#residue.
    def exitResidue(self, ctx: OliviaCSParser.ResidueContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by OliviaCSParser#chemical_shifts.
    def enterChemical_shifts(self, ctx: OliviaCSParser.Chemical_shiftsContext):  # pylint: disable=unused-argument
        self.cur_list_id = max(self.cur_list_id, 0)
        self.cur_list_id += 1

        self.chemShifts = 0

        self.cur_subtype = 'chem_shift'

    # Exit a parse tree produced by OliviaCSParser#chemical_shifts.
    def exitChemical_shifts(self, ctx: OliviaCSParser.Chemical_shiftsContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by OliviaCSParser#chemical_shift.
    def enterChemical_shift(self, ctx: OliviaCSParser.Chemical_shiftContext):  # pylint: disable=unused-argument
        self.chemShifts += 1

    # Exit a parse tree produced by OliviaCSParser#chemical_shift.
    def exitChemical_shift(self, ctx: OliviaCSParser.Chemical_shiftContext):

        try:

            if len(self.anySelection) == 0:
                self.chemShifts -= 1
                return

            index = self.chemShifts
            chain_id = str(ctx.Simple_name(0))
            comp_id = str(ctx.Simple_name(1))
            atom_id = str(ctx.Simple_name(2))
            seq_id = int(str(ctx.Integer()))

            value = self.anySelection[0]
            val_err = self.anySelection[1]

            if value is None:
                self.chemShifts -= 1
                return

            if len(comp_id) == 1 and not self.polyRibonucleotide and not self.polyDeoxyribonucleotide and self.polyPeptide:
                comp_id = next((_comp_id for _comp_id in self.compIdSet if _comp_id in monDict3 and monDict3[_comp_id] == comp_id), comp_id)

            self.atomSelectionSets.clear()

            dstFunc = self.validateCsValue(index, value, val_err)

            if dstFunc is None:
                self.chemShifts -= 1
                return

            _seq_id = seq_id

            auth_seq_id_map = {}

            self.predictSequenceNumberOffsetByFirstResidue(chain_id, _seq_id, comp_id)

            if chain_id in self.offset:
                seq_id += self.offset[chain_id]
                auth_seq_id_map[seq_id] = _seq_id

            L = f'{chain_id}:{seq_id}:{comp_id}:{atom_id}'

            assignment = self.extractAssignment(1, L, index, chain_id, comp_id)

            if assignment is None:
                self.chemShifts -= 1
                return

            has_assignments, has_multiple_assignments = self.checkAssignment(index, assignment)

            self.addCsRow(index, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                          f'{L} -> ')

        finally:
            self.anySelection.clear()

    # Enter a parse tree produced by OliviaCSParser#number.
    def enterNumber(self, ctx: OliviaCSParser.NumberContext):
        if ctx.Float():
            self.anySelection.append(float(str(ctx.Float())))

        elif ctx.Real():
            self.anySelection.append(float(str(ctx.Real())))

        elif ctx.Integer():
            self.anySelection.append(float(str(ctx.Integer())))

        else:
            self.anySelection.append(None)

    # Exit a parse tree produced by OliviaCSParser#number.
    def exitNumber(self, ctx: OliviaCSParser.NumberContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by OliviaCSParser#comment.
    def enterComment(self, ctx: OliviaCSParser.CommentContext):  # pylint: disable=unused-argument
        pass

    # Exit a parse tree produced by OliviaCSParser#comment.
    def exitComment(self, ctx: OliviaCSParser.CommentContext):  # pylint: disable=unused-argument
        pass


# del OliviaCSParser



##
# File: SparkyNPKParserListener.py
# Date: 30-Jan-2025
#
# Updates:
""" ParserLister class for SPARKY PK files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.io.CifReader import CifReader
    from wwpdb.utils.nmr.pk.SparkyNPKParser import SparkyNPKParser
    from wwpdb.utils.nmr.pk.BasePKParserListener import BasePKParserListener
    from wwpdb.utils.nmr.mr.ParserListenerUtil import (REPRESENTATIVE_MODEL_ID,
                                                       REPRESENTATIVE_ALT_ID)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.io.CifReader import CifReader
    from nmr.pk.SparkyNPKParser import SparkyNPKParser
    from nmr.pk.BasePKParserListener import BasePKParserListener
    from nmr.mr.ParserListenerUtil import (REPRESENTATIVE_MODEL_ID,
                                           REPRESENTATIVE_ALT_ID)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator


# This class defines a complete listener for a parse tree produced by SparkyNPKParser.
class SparkyNPKParserListener(ParseTreeListener, BasePKParserListener):

    __has_hz = False
    __has_dev = False
    __has_height = False
    __has_volume = False
    __has_rms = False
    __has_dummy_height = False
    __has_s_n = False
    __has_lw_hz = False
    __has_note = False

    __has_real_vol = False
    __real_vol = None

    __noteSelection = []

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 representativeModelId: int = REPRESENTATIVE_MODEL_ID,
                 representativeAltId: str = REPRESENTATIVE_ALT_ID,
                 mrAtomNameMapping: Optional[List[dict]] = None,
                 cR: Optional[CifReader] = None, caC: Optional[dict] = None, ccU: Optional[ChemCompUtil] = None,
                 csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, representativeModelId, representativeAltId,
                         mrAtomNameMapping, cR, caC, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-pea-spa'
        self.software_name = 'SPARKY'

        if reasons is not None and 'has_real_vol' in reasons:
            self.__has_real_vol = True

    def setInternalMode(self, internal: bool):
        self.internal = internal

    # Enter a parse tree produced by SparkyNPKParser#sparky_npk.
    def enterSparky_npk(self, ctx: SparkyNPKParser.Sparky_npkContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by SparkyNPKParser#sparky_npk.
    def exitSparky_npk(self, ctx: SparkyNPKParser.Sparky_npkContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by SparkyNPKParser#data_label.
    def enterData_label(self, ctx: SparkyNPKParser.Data_labelContext):
        if ctx.W2_LA():
            self.num_of_dim = max(self.num_of_dim, 2)
        if ctx.W3_LA():
            self.num_of_dim = max(self.num_of_dim, 3)
        if ctx.W4_LA():
            self.num_of_dim = max(self.num_of_dim, 4)

        self.__has_hz = False
        if ctx.W2_Hz_LA():
            self.__has_hz = True

        self.__has_dev = False
        if ctx.Dev_w2_LA():
            self.__has_dev = True

        self.__has_height = False
        if ctx.Height_LA():
            self.__has_height = True

        self.__has_volume = False
        if ctx.Volume_LA():
            self.__has_volume = True

        self.__has_rms = False
        if ctx.Dummy_Rms_LA():
            self.__has_rms = True

        self.__has_dummy_height = False
        if ctx.Dummy_H_LA():
            self.__has_dummy_height = True

        self.__has_s_n = False
        if ctx.S_N_LA():
            self.__has_s_n = True

        self.__has_lw_hz = False
        if ctx.Lw2_Hz_LA():
            self.__has_lw_hz = True

        self.__has_note = False
        if ctx.Note_LA():
            self.__has_note = True

        self.initSpectralDim()

    # Exit a parse tree produced by SparkyNPKParser#data_label.
    def exitData_label(self, ctx: SparkyNPKParser.Data_labelContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by SparkyNPKParser#peak_2d.
    def enterPeak_2d(self, ctx: SparkyNPKParser.Peak_2dContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak2d':
            self.num_of_dim = 2
            self.initSpectralDim()

        self.peaks2D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_2d.
    def exitPeak_2d(self, ctx: SparkyNPKParser.Peak_2dContext):

        try:

            index = self.peaks2D

            ass = _ass_ = str(ctx.Assignment_2d_ex())
            if '?' in ass:
                ass = None
            if _ass_ == 'A1H-H' and not self.isFirstResidueAla:
                ass = _ass_ = None

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))

            except ValueError:
                self.peaks2D -= 1
                return

            offset = 0

            x_hz = y_hz = None
            if self.__has_hz:
                if len(self.numberSelection) > offset:
                    x_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_hz = self.numberSelection[offset]
                    offset += 1

            x_dev = y_dev = None
            if self.__has_dev:
                if len(self.numberSelection) > offset:
                    x_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_dev = self.numberSelection[offset]
                    offset += 1

            found_volume = False
            if self.__has_volume and self.__has_real_vol\
               and any(self.originalNumberSelection[_offset] == self.__real_vol for _offset in range(offset, len(self.originalNumberSelection))):
                found_volume = True

            if self.__has_dummy_height and found_volume:
                if not self.__has_height and len(self.originalNumberSelection) > offset:
                    height = self.originalNumberSelection[offset]
                    if height in ("0", "1"):
                        height = None
                offset += 1

            height = volume = None
            if self.__has_height and len(self.originalNumberSelection) > offset:
                height = self.originalNumberSelection[offset]
                offset += 1
            if self.__has_volume and len(self.originalNumberSelection) > offset:
                if self.__has_real_vol:
                    volume = self.__real_vol
                    if volume == self.originalNumberSelection[offset]:
                        offset += 1
                else:
                    volume = self.originalNumberSelection[offset]
                    offset += 1

            if self.__has_rms and found_volume:
                offset += 1

            if self.__has_s_n:
                offset += 1

            x_lw_hz = y_lw_hz = None
            if self.__has_lw_hz:
                if len(self.numberSelection) > offset:
                    x_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_lw_hz = self.numberSelection[offset]
                    offset += 1

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm):
                self.peaks2D -= 1
                return

            dstFunc = self.validatePeak2D(index, x_ppm, y_ppm, x_dev, y_dev, None, None,
                                          x_hz, y_hz, x_lw_hz, y_lw_hz, height, None, volume, None)

            if dstFunc is None:
                self.peaks2D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2 =\
                    self.checkAssignments2D(index, assignments, dstFunc)

                if not has_assignments and any(a[0]['comp_id'] == 'ARG' and a[0]['atom_id'] == 'NH' for a in assignments if a is not None):
                    _ass_ = ass.replace('NH', 'HN')
                    assignments = []
                    hint = None
                    for _ass in _ass_.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None

                    has_assignments, has_multiple_assignments, asis1, asis2 =\
                        self.checkAssignments2D(index, assignments, dstFunc)

            self.addAssignedPkRow2D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ == '?-?' else _ass_)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__real_vol = None
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#peak_3d.
    def enterPeak_3d(self, ctx: SparkyNPKParser.Peak_3dContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak3d':
            self.num_of_dim = 3
            self.initSpectralDim()

        self.peaks3D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_3d.
    def exitPeak_3d(self, ctx: SparkyNPKParser.Peak_3dContext):

        try:

            index = self.peaks3D

            ass = _ass_ = str(ctx.Assignment_3d_ex())
            if '?' in ass:
                ass = None
            if _ass_ == 'A1H-H-H' and not self.isFirstResidueAla:
                ass = _ass_ = None

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))
                z_ppm = float(str(ctx.Float(2)))

            except ValueError:
                self.peaks3D -= 1
                return

            offset = 0

            x_hz = y_hz = z_hz = None
            if self.__has_hz:
                if len(self.numberSelection) > offset:
                    x_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_hz = self.numberSelection[offset]
                    offset += 1

            x_dev = y_dev = z_dev = None
            if self.__has_dev:
                if len(self.numberSelection) > offset:
                    x_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_dev = self.numberSelection[offset]
                    offset += 1

            found_volume = False
            if self.__has_volume and self.__has_real_vol\
               and any(self.originalNumberSelection[_offset] == self.__real_vol for _offset in range(offset, len(self.originalNumberSelection))):
                found_volume = True

            if self.__has_dummy_height and found_volume:
                if not self.__has_height and len(self.originalNumberSelection) > offset:
                    height = self.originalNumberSelection[offset]
                    if height in ("0", "1"):
                        height = None
                offset += 1

            height = volume = None
            if self.__has_height and len(self.originalNumberSelection) > offset:
                height = self.originalNumberSelection[offset]
                offset += 1
            if self.__has_volume and len(self.originalNumberSelection) > offset:
                if self.__has_real_vol:
                    volume = self.__real_vol
                    if volume == self.originalNumberSelection[offset]:
                        offset += 1
                else:
                    volume = self.originalNumberSelection[offset]
                    offset += 1

            if self.__has_rms and found_volume:
                offset += 1

            if self.__has_s_n:
                offset += 1

            x_lw_hz = y_lw_hz = z_lw_hz = None
            if self.__has_lw_hz:
                if len(self.numberSelection) > offset:
                    x_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_lw_hz = self.numberSelection[offset]
                    offset += 1

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm):
                self.peaks3D -= 1
                return

            dstFunc = self.validatePeak3D(index, x_ppm, y_ppm, z_ppm, x_dev, y_dev, z_dev, None, None, None,
                                          x_hz, y_hz, z_hz, x_lw_hz, y_lw_hz, z_lw_hz, height, None, volume, None)

            if dstFunc is None:
                self.peaks3D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                    self.checkAssignments3D(index, assignments, dstFunc)

                if not has_assignments and any(a[0]['comp_id'] == 'ARG' and a[0]['atom_id'] == 'NH' for a in assignments if a is not None):
                    _ass_ = ass.replace('NH', 'HN')
                    assignments = []
                    hint = None
                    for _ass in _ass_.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None

                    has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                        self.checkAssignments3D(index, assignments, dstFunc)

            self.addAssignedPkRow3D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ == '?-?-?' else _ass_)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__real_vol = None
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#peak_4d.
    def enterPeak_4d(self, ctx: SparkyNPKParser.Peak_4dContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak4d':
            self.num_of_dim = 4
            self.initSpectralDim()

        self.peaks4D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_4d.
    def exitPeak_4d(self, ctx: SparkyNPKParser.Peak_4dContext):

        try:

            index = self.peaks4D

            ass = _ass_ = str(ctx.Assignment_4d_ex())
            if '?' in ass:
                ass = None
            if _ass_ == 'A1H-H-H-H' and not self.isFirstResidueAla:
                ass = _ass_ = None

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))
                z_ppm = float(str(ctx.Float(2)))
                a_ppm = float(str(ctx.Float(3)))

            except ValueError:
                self.peaks4D -= 1
                return

            offset = 0

            x_hz = y_hz = z_hz = a_hz = None
            if self.__has_hz:
                if len(self.numberSelection) > offset:
                    x_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    a_hz = self.numberSelection[offset]
                    offset += 1

            x_dev = y_dev = z_dev = a_dev = None
            if self.__has_dev:
                if len(self.numberSelection) > offset:
                    x_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_dev = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    a_dev = self.numberSelection[offset]
                    offset += 1

            found_volume = False
            if self.__has_volume and self.__has_real_vol\
               and any(self.originalNumberSelection[_offset] == self.__real_vol for _offset in range(offset, len(self.originalNumberSelection))):
                found_volume = True

            if self.__has_dummy_height and found_volume:
                if not self.__has_height and len(self.originalNumberSelection) > offset:
                    height = self.originalNumberSelection[offset]
                    if height in ("0", "1"):
                        height = None
                offset += 1

            height = volume = None
            if self.__has_height and len(self.originalNumberSelection) > offset:
                height = self.originalNumberSelection[offset]
                offset += 1
            if self.__has_volume and len(self.originalNumberSelection) > offset:
                if self.__has_real_vol:
                    volume = self.__real_vol
                    if volume == self.originalNumberSelection[offset]:
                        offset += 1
                else:
                    volume = self.originalNumberSelection[offset]
                    offset += 1

            if self.__has_rms and found_volume:
                offset += 1

            if self.__has_s_n:
                offset += 1

            x_lw_hz = y_lw_hz = z_lw_hz = a_lw_hz = None
            if self.__has_lw_hz:
                if len(self.numberSelection) > offset:
                    x_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    y_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    z_lw_hz = self.numberSelection[offset]
                    offset += 1
                if len(self.numberSelection) > offset:
                    a_lw_hz = self.numberSelection[offset]
                    offset += 1

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm, a_ppm):
                self.peaks4D -= 1
                return

            dstFunc = self.validatePeak4D(index, x_ppm, y_ppm, z_ppm, a_ppm, x_dev, y_dev, z_dev, a_dev, None, None, None, None,
                                          x_hz, y_hz, z_hz, a_hz, x_lw_hz, y_lw_hz, z_lw_hz, a_lw_hz, height, None, volume, None)

            if dstFunc is None:
                self.peaks4D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)
            cur_spectral_dim[4]['freq_hint'].append(a_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = asis4 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                    self.checkAssignments4D(index, assignments, dstFunc)

                if not has_assignments and any(a[0]['comp_id'] == 'ARG' and a[0]['atom_id'] == 'NH' for a in assignments if a is not None):
                    _ass_ = ass.replace('NH', 'HN')
                    assignments = []
                    hint = None
                    for _ass in _ass_.split('-'):
                        assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                        hint = assignments[-1] if assignments[-1] is not None else None

                    has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                        self.checkAssignments4D(index, assignments, dstFunc)

            self.addAssignedPkRow4D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3, asis4,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ == '?-?-?-?' else _ass_)

        finally:
            self.numberSelection.clear()
            self.originalNumberSelection.clear()
            self.__real_vol = None
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#peak_2d_po.
    def enterPeak_2d_po(self, ctx: SparkyNPKParser.Peak_2d_poContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak2d':
            self.num_of_dim = 2
            self.initSpectralDim()

        self.peaks2D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_2d_po.
    def exitPeak_2d_po(self, ctx: SparkyNPKParser.Peak_2d_poContext):

        try:

            index = self.peaks2D

            ass = _ass_ = None
            if ctx.Assignment_2d_ex():
                ass = _ass_ = str(ctx.Assignment_2d_ex())
                if '?' in ass:
                    ass = None
                if _ass_ == 'A1H-H' and not self.isFirstResidueAla:
                    ass = _ass_ = None
            elif ctx.Assignment_3d_ex():
                _ass_ = str(ctx.Assignment_3d_ex())
            elif ctx.Assignment_4d_ex():
                _ass_ = str(ctx.Assignment_4d_ex())
            if _ass_ in ('?-?', None) and self.__has_note and len(self.__noteSelection) > 0:
                _ass_ = ' '.join(self.__noteSelection)

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))

            except ValueError:
                self.peaks2D -= 1
                return

            if None in (x_ppm, y_ppm):
                self.peaks2D -= 1
                return

            dstFunc = self.validatePeak2D(index, x_ppm, y_ppm, None, None, None, None,
                                          None, None, None, None, None, None, None, None)

            if dstFunc is None:
                self.peaks2D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2 =\
                    self.checkAssignments2D(index, assignments, dstFunc)

            self.addAssignedPkRow2D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ in ('?-?', None) else _ass_)

        finally:
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#peak_3d_po.
    def enterPeak_3d_po(self, ctx: SparkyNPKParser.Peak_3d_poContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak3d':
            self.num_of_dim = 3
            self.initSpectralDim()

        self.peaks3D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_3d_po.
    def exitPeak_3d_po(self, ctx: SparkyNPKParser.Peak_3d_poContext):

        try:

            index = self.peaks3D

            ass = _ass_ = None
            if ctx.Assignment_3d_ex():
                ass = _ass_ = str(ctx.Assignment_3d_ex())
                if '?' in ass:
                    ass = None
                if _ass_ == 'A1H-H-H' and not self.isFirstResidueAla:
                    ass = _ass_ = None
            if ctx.Assignment_4d_ex():
                _ass_ = str(ctx.Assignment_4d_ex())
            if _ass_ in ('?-?-?', None) and self.__has_note and len(self.__noteSelection) > 0:
                _ass_ = ' '.join(self.__noteSelection)

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))
                z_ppm = float(str(ctx.Float(2)))

            except ValueError:
                self.peaks3D -= 1
                return

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm):
                self.peaks3D -= 1
                return

            dstFunc = self.validatePeak3D(index, x_ppm, y_ppm, z_ppm, None, None, None, None, None, None,
                                          None, None, None, None, None, None, None, None, None, None)

            if dstFunc is None:
                self.peaks3D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2, asis3 =\
                    self.checkAssignments3D(index, assignments, dstFunc)

            self.addAssignedPkRow3D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ in ('?-?-?', None) else _ass_)

        finally:
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#peak_4d_po.
    def enterPeak_4d_po(self, ctx: SparkyNPKParser.Peak_4d_poContext):  # pylint: disable=unused-argument
        if self.cur_subtype != 'peak4d':
            self.num_of_dim = 4
            self.initSpectralDim()

        self.peaks4D += 1

        self.atomSelectionSets.clear()
        self.asIsSets.clear()

    # Exit a parse tree produced by SparkyNPKParser#peak_4d_po.
    def exitPeak_4d_po(self, ctx: SparkyNPKParser.Peak_4d_poContext):

        try:

            index = self.peaks4D

            ass = _ass_ = None
            if ctx.Assignment_4d_ex():
                ass = _ass_ = str(ctx.Assignment_4d_ex())
                if '?' in ass:
                    ass = None
                if _ass_ == 'A1H-H-H-H' and not self.isFirstResidueAla:
                    ass = _ass_ = None
            if _ass_ in ('?-?-?-?', None) and self.__has_note and len(self.__noteSelection) > 0:
                _ass_ = ' '.join(self.__noteSelection)

            try:

                x_ppm = float(str(ctx.Float(0)))
                y_ppm = float(str(ctx.Float(1)))
                z_ppm = float(str(ctx.Float(2)))
                a_ppm = float(str(ctx.Float(3)))

            except ValueError:
                self.peaks4D -= 1
                return

            if not self.hasPolySeq and not self.hasNonPolySeq:
                return

            if None in (x_ppm, y_ppm, z_ppm, a_ppm):
                self.peaks4D -= 1
                return

            dstFunc = self.validatePeak4D(index, x_ppm, y_ppm, z_ppm, a_ppm, None, None, None, None, None, None, None, None,
                                          None, None, None, None, None, None, None, None, None, None, None, None)

            if dstFunc is None:
                self.peaks4D -= 1
                return

            cur_spectral_dim = self.spectral_dim[self.num_of_dim][self.cur_list_id]

            cur_spectral_dim[1]['freq_hint'].append(x_ppm)
            cur_spectral_dim[2]['freq_hint'].append(y_ppm)
            cur_spectral_dim[3]['freq_hint'].append(z_ppm)
            cur_spectral_dim[4]['freq_hint'].append(a_ppm)

            has_assignments = has_multiple_assignments = False
            asis1 = asis2 = asis3 = asis4 = None

            if ass is not None:
                assignments = []
                hint = None
                for _ass in ass.split('-'):
                    assignments.append(self.extractPeakAssignment(1, _ass, index, hint=hint))
                    hint = assignments[-1] if assignments[-1] is not None else None

                has_assignments, has_multiple_assignments, asis1, asis2, asis3, asis4 =\
                    self.checkAssignments4D(index, assignments, dstFunc)

            self.addAssignedPkRow4D(index, dstFunc, has_assignments, has_multiple_assignments,
                                    asis1, asis2, asis3, asis4,
                                    f'{ass} -> ',
                                    None if (has_assignments and not has_multiple_assignments)
                                    or _ass_ in ('?-?-?-?', None) else _ass_)

        finally:
            self.__noteSelection.clear()

    # Enter a parse tree produced by SparkyNPKParser#number.
    def enterNumber(self, ctx: SparkyNPKParser.NumberContext):

        try:

            if ctx.Float():
                value = str(ctx.Float())
                self.numberSelection.append(float(value))
                self.originalNumberSelection.append(value)

            elif ctx.Real():
                value = str(ctx.Real())
                self.numberSelection.append(float(value))
                self.originalNumberSelection.append(value)

            elif ctx.Real_vol():
                self.__has_real_vol = True
                if self.reasons is None and 'has_real_vol' not in self.reasonsForReParsing:
                    self.reasonsForReParsing['has_real_vol'] = True
                value = str(ctx.Real_vol())
                if ' ' in value:
                    value = value.split()[0]
                    self.__real_vol = value
                self.numberSelection.append(float(value))
                self.originalNumberSelection.append(value)

            else:
                value = str(ctx.Integer())
                self.numberSelection.append(int(value))
                self.originalNumberSelection.append(value)

        except ValueError:
            self.numberSelection.append(None)
            self.originalNumberSelection.append(None)

    # Exit a parse tree produced by SparkyNPKParser#number.
    def exitNumber(self, ctx: SparkyNPKParser.NumberContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by SparkyNPKParser#note.
    def enterNote(self, ctx: SparkyNPKParser.NoteContext):
        if ctx.Simple_name():
            self.__noteSelection.append(str(ctx.Simple_name()))

        elif ctx.Float():
            self.__noteSelection.append(str(ctx.Float()))

        elif ctx.Integer():
            self.__noteSelection.append(str(ctx.Integer()))

        elif ctx.Note_2d_ex():
            self.__noteSelection.append(str(ctx.Note_2d_ex()))

        elif ctx.Note_3d_ex():
            self.__noteSelection.append(str(ctx.Note_3d_ex()))

        elif ctx.Note_4d_ex():
            self.__noteSelection.append(str(ctx.Note_4d_ex()))

    # Exit a parse tree produced by SparkyNPKParser#note.
    def exitNote(self, ctx: SparkyNPKParser.NoteContext):  # pylint: disable=unused-argument
        pass


# del SparkyNPKParser
